(*                      }
{*  MODULE:  stdmodh.p   }
{*  module with PASCAL-SC standard procedures, functions and operators }
{*   for hardware arithmetic including directed roundings }
{*  last modification:  29.01.91  09:06   }
{* uli 29.01.91 : a_ival, r_rval          }
{*)

module stdmod ;
procedure roundnear ; external r_rndn ;
procedure rounddown ; external r_rndd ;
procedure roundup   ; external r_rndu ;
procedure roundchop ; external r_rndt ;
procedure setround (i:integer) ; external r_srnd ;

global function abs (  i : integer) : integer ; external a_abs_ ;
global function sqr (  i : integer) : integer ; external a_sqr_ ;
global operator div (i,j : integer)r: integer ; external a_div_ ;
global operator mod (i,j : integer)r: integer ; external a_mod_ ;
global operator /   (i,j : integer)r: real ;
	 begin r:= real(i) / real(j) end ;
global function trunc   (r : real ) : integer ; external f_trun ;
global function round   (r : real ) : integer ; external f_rond ;
global function abs     (r : real ) : real    ; external fabs ;
global function sqr     (r : real ) : real    ; begin sqr := r*r end ;
global function sqrt    (r : real ) : real    ; external sqrt ;

global function exp  (x : real)              : real ; external exp  ;
global function ln   (x : real)              : real ; external log  ;
global function sin  (x : real)              : real ; external sin  ;
global function cos  (x : real)              : real ; external cos  ;
global function tan  (x : real)              : real ; external tan  ;
global function arctan (x : real)            : real ; external atan ;

global operator +< (x,y : real) r : real ;
  begin
    rounddown ;
    r := x+y ;
    roundnear ;
  end ;
global operator +> (x,y : real) r : real ;
  begin
    roundup ;
    r := x+y ;
    roundnear ;
  end ;
global operator -< (x,y : real) r : real ;
  begin
    rounddown ;
    r := x-y ;
    roundnear ;
  end ;
global operator -> (x,y : real) r : real ;
  begin
    roundup ;
    r := x-y ;
    roundnear ;
  end ;
global operator *< (x,y : real) r : real ;
  begin
    rounddown ;
    r := x*y ;
    roundnear ;
  end ;
global operator *> (x,y : real) r : real ;
  begin
    roundup ;
    r := x*y ;
    roundnear ;
  end ;
global operator /< (x,y : real) r : real ;
  begin
    rounddown ;
    r := x/y ;
    roundnear ;
  end ;
global operator /> (x,y : real) r : real ;
  begin
    roundup ;
    r := x/y ;
    roundnear ;
  end ;

{   string subroutines   }

global function string (c:char)       : string ;  external s_char ;
global operator +    (x,y : string) r : string ;  external s_conc ;
global operator := (var x : string; y : string) ; external s_asgn ;

global operator =  ( s,t : string )         sseq : boolean;  external s_sseq;
global operator >= ( s,t : string )         ssge : boolean;  external s_ssge;
global operator >  ( s,t : string )         ssgt : boolean;  external s_ssgt;
global operator <= ( s,t : string )         ssle : boolean;  external s_ssle;
global operator <  ( s,t : string )         sslt : boolean;  external s_sslt;
global operator <> ( s,t : string )         ssne : boolean;  external s_ssne;
global operator =  ( s : string; t : char ) sceq : boolean;  external s_sceq;
global operator >= ( s : string; t : char ) scge : boolean;  external s_scge;
global operator >  ( s : string; t : char ) scgt : boolean;  external s_scgt;
global operator <= ( s : string; t : char ) scle : boolean;  external s_scle;
global operator <  ( s : string; t : char ) sclt : boolean;  external s_sclt;
global operator <> ( s : string; t : char ) scne : boolean;  external s_scne;
global operator =  ( s : char; t : string ) cseq : boolean;  external s_cceq;
global operator >= ( s : char; t : string ) csge : boolean;  external s_csge;
global operator >  ( s : char; t : string ) csgt : boolean;  external s_csgt;
global operator <= ( s : char; t : string ) csle : boolean;  external s_csle;
global operator <  ( s : char; t : string ) cslt : boolean;  external s_cslt;
global operator <> ( s : char; t : string ) csne : boolean;  external s_csne;

global function     length( s : string )           :integer ; external s_len_;
global function  maxlength( s : string )           :integer ; external s_alen;
global procedure setlength( var s : string; n:integer)      ; external s_slen;
global function     pos   (s,      t: string )     :integer ; external s_spos;
global function     pos   (c:char; s: string )     :integer ; external s_cpos;
global function substring (s: string; i,n: integer): string ; external s_subs;

global function ival ( s : string; var rest: string  ) : integer ;
      external a_sval;
global function ival ( s : string                    ) : integer ;
      external a_ival;
global function rval ( s : string; rnd : integer; var rest: string) : real ;
      external r_sval;
global function rval ( s : string; rnd : integer) : real ;
      external r_rval;
global function rval ( s : string; var rest: string) : real ;
		      begin rval := rval (s,0,rest) end ;
global function rval ( s : string) : real ;
  var dummy:string;   begin rval := rval (s,0,dummy) end ;
global function image (i:integer; w : integer ) : string;
      external s_int_;
global function image (i:integer) : string;
		      begin image := image (i,0) end ;
global function image ( r : real; w,f,rnd : integer ) : string;
      external s_real;
global function image ( r : real; w,f     : integer ) : string;
		      begin image := image (r,w,f,0) end ;
global function image ( r : real; w       : integer ) : string;
		      begin image := image (r,w,0,0) end ;
global function image ( r : real                    ) : string;
		      begin image := image (r,0,0,0) end ;

{ PASCAL-XSC standard input/output }

global procedure readln  (var t : text) ; external f_rdln ;
global procedure writeln (var t : text) ; external f_wrln ;
global procedure read  (var t : text; var c : char              ) ; external f_rdh1 ;
global procedure read ( var t : text; var s : string            ) ; external f_rds1;
global procedure read  (var t : text; var i : integer           ) ; external f_rdi1 ;
global procedure read  (var t : text; var r : real              ) ; external f_rdf1 ;
global procedure read  (var t : text; var r : real; i : integer ) ; external f_rdf2 ;

global procedure write (var t : text; c : char                ) ; external f_wrh1 ;
global procedure write (var t : text; c : char   ; i : integer) ; external f_wrh2 ;
global procedure write( var t : text; s : string              ) ; external f_wrs1;
global procedure write( var t : text; s : string ; i : integer) ; external f_wrs2;
 
global procedure write (var t : text; i : integer             ) ; external f_wri1 ;
global procedure write (var t : text; i ,          j : integer) ; external f_wri2 ;
global procedure write (var t : text; b : boolean             ) ; external f_wrb1 ;
global procedure write (var t : text; b : boolean; i : integer) ; external f_wrb2 ;
global procedure write (var t : text; r : real                ) ; external f_wrf1 ;
global procedure write (var t : text; r : real; i    : integer) ; external f_wrf2 ;
global procedure write (var t : text; r : real; i, j : integer) ; external f_wrf3 ;
global procedure write (var t : text; r : real; i,j,k: integer) ; external f_wrf4 ;

global procedure page  (var t : text) ; begin write (t, chr(12)) end ;
global procedure page ; begin write (chr(12)) end ;

global procedure reset   (var t : text ;  s : string ) ; external f_srse ;
global procedure rewrite (var t : text ;  s : string ) ; external f_srwi ;

end.
